"""

Abstract class for data augmentation.

It was inspired by nlpaug (https://nlpaug.readthedocs.io/en/latest/augmenter/word/spelling.html)

These class allow to augmentate data with different ways: with spelling, language generative models, masked language models, back translation augmenter, word embeddings, abstractive summarization and synonyms.

TO DO:
* DataAugmentationSynonyms
** __init__(method, stopwords, aug_percent, dictionary, top_k)
* DataAugmentationWordEmbeddings
** __init__(method, stopwords, aug_percent, model, device, top_k)
* DataAugmentationGenerativeOpenSource
** __init__(method, stopwords, max_words, model, device, temperature, top_k, top_p, device, frequency_penalty, presence_penalty)
* DataAugmentationGenerativeOpenAI
** __init__(method, stopwords, max_words, model, token, temperature, top_k, top_p, frequency_penalty, presence_penalty)
* DataAugmentationBackTranslation
** __init__(method, stopwords, max_words, model, device, temperature, top_k)
* DataAugmentationAbstractiveSummarization
** __init__(method, stopwords, max_words, model, device, temperature, top_k)

DataAugmentationSynonyms uses WordNet to find synonyms.
DataAugmentationWordEmbeddings uses GenSim to find similar words (word2vec or fasttext)
Masked, DataAugmentationGenerative, DataAugmentationBackTranslation and DataAugmentationAbstractiveSummarization use HuggingFace transformers.

"""

import os

import es_core_news_sm
import pandas as pd
import swifter

from datasets import load_dataset
from tqdm import tqdm
tqdm.pandas()

class DataAugmentationAbstract:
    """
    Abstract class for data augmentation.

    """

    def __init__(self, method):
        """
        Initialize the class.
        """
        self.method = method

    def augment(self, texts, num_samples=1, num_workers=-1):
        """
        Augment the data.
        """
        if num_workers == -1:
            # Get the number of CPUs in the system
            num_workers = os.cpu_count() - 1
        if isinstance(texts, str):
            return self._text_augment_(texts, num_samples)
        elif isinstance(texts, list):
            return self._list_augment_(texts, num_samples, num_workers)
        elif isinstance(texts, pd.Series):
            return self._pandas_augment_(texts, num_samples, num_workers)
        elif isinstance(texts, load_dataset.Dataset):
            return self._datasets_augment_(texts, num_samples, num_workers)
        else:
            raise ValueError(
                "The texts must be a string, a list of strings or a pandas Series."
            )

    def _text_augment_(self, text, num_sample):
        """
        Augment a single text.
        """
        raise NotImplementedError("This method must be implemented in the child class.")

    def _list_augment_(self, texts, num_samples, num_workers):
        """
        Augment a list of texts.
        """
        return self._pandas_augment_(
            pd.Series(texts), num_samples, num_workers
        ).tolist()

    def _pandas_augment_(self, texts, num_samples, num_workers):
        """
        Augment a pandas Series.
        """
        # return texts.progress_apply(
        #     self._text_augment_,
        # )
        if num_workers == 1:
            return texts.progress_apply(
                self._text_augment_,
                num_samples=num_samples
            )
        else:
            return texts.swifter.apply(
                self._text_augment_,
                num_samples=num_samples,
            )

    def _datasets_augment_(self, dataset, num_samples, num_workers):
        """
        Augment a datasets Dataset.
        """
        # Apply self._text_augment_(text) to every text in dataset["text"] (HuggingFace datasets)
        dataset = dataset.map(
            self._text_augment_,
            batched=True,
            num_proc=num_workers,
            load_from_cache_file=False,
        )
        return dataset


    def _load_default_tokenizer_(self):
        # import es_core_news_sm if it is not imported
        # if "es_core_news_sm" not in sys.modules:
        #     import es_core_news_sm

        if not hasattr(self, "nlp_spacy"):
            self.nlp_spacy = es_core_news_sm.load(
                disable=["ner", "parser", "tagger", "textcat", "vectors"]
            )

        def tokenizer(text):
            return [token.text for token in self.nlp_spacy(text)]

        self.tokenizer = tokenizer

    def _load_default_stopwords_(self):
        from spanish_nlp.utils.stopwords import extended_stopwords

        self.stopwords = extended_stopwords

    def set_stopwords(self, stopwords):
        """
        Set the stopwords.
        """
        self.stopwords = stopwords



# -*- coding: UTF-8 -*-

"""
Emoticons and Emoji data dictonary
"""

EMOTICONS = {
    ":‑)": "happy_face_or_smiley_1",
    ":)": "happy_face_or_smiley_2",
    ":-]": "happy_face_or_smiley_3",
    ":]": "happy_face_or_smiley_4",
    ":-3": "happy_face_smiley_5",
    ":3": "happy_face_smiley_6",
    ":->": "happy_face_smiley_7",
    ":>": "happy_face_smiley_8",
    "8-)": "happy_face_smiley_9",
    ":o)": "happy_face_smiley_10",
    ":-}": "happy_face_smiley_11",
    ":}": "happy_face_smiley_12",
    ":-)": "happy_face_smiley_13",
    "C:": "happy_face_smiley_14",
    ":^)": "happy_face_smiley_15",
    "=]": "happy_face_smiley_16",
    "=)": "happy_face_smiley_17",
    "c:": "happy_face_smiley_14",
    ":‑D": "laughing_big_grin_or_laugh_with_glasses_1",
    ":D": "laughing_big_grin_or_laugh_with_glasses_2",
    "8‑D": "laughing_big_grin_or_laugh_with_glasses_3",
    "8D": "laughing_big_grin_or_laugh_with_glasses_4",
    "X‑D": "laughing_big_grin_or_laugh_with_glasses_5",
    "XD": "laughing_big_grin_or_laugh_with_glasses_6",
    "=D": "laughing_big_grin_or_laugh_with_glasses_7",
    "xD": "laughing_big_grin_or_laugh_with_glasses_8",
    "=3": "laughing_big_grin_or_laugh_with_glasses_9",
    "B^D": "laughing_big_grin_or_laugh_with_glasses_10",
    ":-))": "very_happy",
    ":‑(": "frown_sad_andry_or_pouting_1",
    ":-(": "frown_sad_andry_or_pouting_2",
    ":(": "frown_sad_andry_or_pouting_3",
    ":‑c": "frown_sad_andry_or_pouting_4",
    ":c": "frown_sad_andry_or_pouting_5",
    ":‑<": "frown_sad_andry_or_pouting_6",
    ":<": "frown_sad_andry_or_pouting_7",
    ":‑[": "frown_sad_andry_or_pouting_8",
    ":[": "frown_sad_andry_or_pouting_9",
    ":-||": "frown_sad_andry_or_pouting_10",
    ">:[": "frown_sad_andry_or_pouting_11",
    ":{": "frown_sad_andry_or_pouting_12",
    ":@": "frown_sad_andry_or_pouting_13",
    ">:(": "frown_sad_andry_or_pouting_14",
    ":'‑(": "crying_1",
    ":'(": "crying_3",
    ":'‑)": "tears_of_happiness_1",
    ":')": "tears_of_happiness_2",
    "D‑':": "horror",
    "D:<": "disgust",
    "D:": "sadness",
    "D8": "great_dismay_1",
    "D;": "great_dismay_2",
    "D=": "great_dismay_3",
    "DX": "great_dismay_4",
    ":‑O": "surprise_5",
    ":O": "surprise_6",
    ":‑o": "surprise_7",
    ":o": "surprise_8",
    ":-0": "shock",
    "8‑0": "yawn_1",
    ">:O": "yawn_2",
    ":-*": "kiss",
    ":*": "kiss_2",
    ":X": "kiss_3",
    ";‑)": "wink_or_smirk_1",
    ";)": "wink_or_smirk_2",
    "*-)": "wink_or_smirk_3",
    "*)": "wink_or_smirk_4",
    ";‑]": "wink_or_smirk_5",
    ";]": "wink_or_smirk_6",
    ";^)": "wink_or_smirk_7",
    ":‑,": "wink_or_smirk_8",
    ";D": "wink_or_smirk_9",
    ":‑P": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_1",
    ":P": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_2",
    "X‑P": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_3",
    "XP": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_4",
    ":‑Þ": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_5",
    ":Þ": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_6",
    ":b": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_7",
    "d:": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_8",
    "=p": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_9",
    ">:P": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry_10",
    ":‑/": "skeptical_annoyed_undecided_uneasy_or_hesitant_1",
    ":/": "skeptical_annoyed_undecided_uneasy_or_hesitant_2",
    ":-[.]": "skeptical_annoyed_undecided_uneasy_or_hesitant_3",
    ">:[()]": "skeptical_annoyed_undecided_uneasy_or_hesitant_4",
    ">:/": "skeptical_annoyed_undecided_uneasy_or_hesitant_5",
    ":[()]": "skeptical_annoyed_undecided_uneasy_or_hesitant_6",
    "=/": "skeptical_annoyed_undecided_uneasy_or_hesitant_7",
    "=[()]": "skeptical_annoyed_undecided_uneasy_or_hesitant_8",
    ":L": "skeptical_annoyed_undecided_uneasy_or_hesitant_9",
    "=L": "skeptical_annoyed_undecided_uneasy_or_hesitant_10",
    ":S": "skeptical_annoyed_undecided_uneasy_or_hesitant_11",
    ":‑|": "straight_face_1",
    ":|": "straight_face_2",
    ":$": "embarrassed_or_blushing_1",
    ":-$": "embarrassed_or_blushing_2",
    ":‑x": "sealed_lips_or_wearing_braces_or_tongue-tied_1",
    ":x": "sealed_lips_or_wearing_braces_or_tongue-tied_2",
    ":‑#": "sealed_lips_or_wearing_braces_or_tongue-tied_3",
    ":#": "sealed_lips_or_wearing_braces_or_tongue-tied_4",
    ":‑&": "sealed_lips_or_wearing_braces_or_tongue-tied_5",
    ":&": "sealed_lips_or_wearing_braces_or_tongue-tied_6",
    "O:‑)": "angel_saint_or_innocent_1",
    "O:)": "angel_saint_or_innocent_2",
    "0:‑3": "angel_saint_or_innocent_3",
    "0:3": "angel_saint_or_innocent_4",
    "0:‑)": "angel_saint_or_innocent_5",
    "0:)": "angel_saint_or_innocent_6",
    ":‑b": "tongue_sticking_out_cheeky_playful_or_blowing_a_raspberry",
    "0;^)": "angel_saint_or_innocent",
    ">:‑)": "evil_or_devilish_1",
    ">:)": "evil_or_devilish_2",
    "}:‑)": "evil_or_devilish_3",
    "}:)": "evil_or_devilish_4",
    "3:‑)": "evil_or_devilish_5",
    "3:)": "evil_or_devilish_6",
    ">;)": "evil_or_devilish_7",
    "|;‑)": "cool",
    "|‑O": "bored",
    ":‑J": "tongue-in-cheek",
    "#‑)": "party_all_night",
    "%‑)": "drunk_or_confused_1",
    "%)": "drunk_or_confused_2",
    ":-###..": "being_sick_1",
    ":###..": "being_sick_2",
    "<:‑|": "dump",
    "(>_<)": "troubled_1",
    "(>_<)>": "troubled_2",
    "(';')": "baby",
    "(^^>``": "nervous_or_embarrassed_or_troubled_or_shy_or_sweat_drop_1",
    "(^_^;)": "nervous_or_embarrassed_or_troubled_or_shy_or_sweat_drop_2",
    "(-_-;)": "nervous_or_embarrassed_or_troubled_or_shy_or_sweat_drop_3",
    "(~_~;) (・.・;)": "nervous_or_embarrassed_or_troubled_or_shy_or_sweat_drop_4",
    "(-_-)zzz": "sleeping",
    "(^_-)": "wink",
    "(o|o)": "ultraman",
    "^_^": "joyful_1",
    "(^_^)/": "joyful_2",
    "(^O^)／": "joyful_3",
    "(^o^)／": "joyful_4",
    "(__)": "kowtow_as_a_sign_of_respect_or_dogeza_for_apology_1",
    "_(._.)_": "kowtow_as_a_sign_of_respect_or_dogeza_for_apology_2",
    "<(_ _)>": "kowtow_as_a_sign_of_respect_or_dogeza_for_apology_3",
    "<m(__)m>": "kowtow_as_a_sign_of_respect_or_dogeza_for_apology_4",
    "m(__)m": "kowtow_as_a_sign_of_respect_or_dogeza_for_apology_5",
    "m(_ _)m": "kowtow_as_a_sign_of_respect_or_dogeza_for_apology_5",
    "('_')": "sad_or_crying_1",
    "(/_;)": "sad_or_crying_2",
    "(T_T) (;_;)": "sad_or_crying_3",
    "(;_;": "sad_of_crying_4",
    "(;_:)": "sad_or_crying_5",
    "(;O;)": "sad_or_crying_6",
    "(:_;)": "sad_or_crying_7",
    "(ToT)": "sad_or_crying_8",
    ";_;": "sad_or_crying_1",
    ";-;": "sad_or_crying_2",
    ";n;": "sad_or_crying_3",
    "Q.Q": "sad_or_crying_4",
    "T.T": "sad_or_crying_5",
    "T_T": "sad_or_crying_6",
    "QQ": "sad_or_crying_7",
    "Q_Q": "sad_or_crying_8",
    "(-.-)": "shame_1",
    "(-_-)": "shame_2",
    "(一一)": "shame_3",
    "(；一_一)": "shame_4",
    "(=_=)": "tired_5",
    "(=^·^=)": "cat_1",
    "(=^··^=)": "cat_2",
    "=_^=\t": "cat_3",
    "(..)": "looking_down_1",
    "(._.)": "looking_down_2",
    "^m^": "giggling_with_hand_covering_mouth",
    ">^_^<": "normal_laugh_1",
    "<^!^>": "normal_laugh_2",
    "^/^": "normal_laugh_3",
    "（*^_^*）": "normal_laugh_4",
    "(^<^) (^.^)": "normal_laugh_5",
    "(^^)": "normal_laugh_10",
    "(^.^)": "normal_laugh_7",
    "(^_^.)": "normal_laugh_8",
    "(^_^)": "normal_laugh_9",
    "(^J^)": "normal_laugh_11",
    "(*^.^*)": "normal_laugh_12",
    "(^—^）": "normal_laugh_13",
    "(#^.^#)": "normal_laugh_14",
    "（^—^）": "waving_1",
    "(;_;)/~~~": "waving_2",
    "(^.^)/~~~": "waving_3",
    "(-_-)/~~~ ($··)/~~~": "waving_4",
    "(T_T)/~~~": "waving_5",
    "(ToT)/~~~": "waving_6",
    "(*^0^*)": "excited",
    "(*_*)": "amazed_1",
    "(*_*;": "amazed_2",
    "(+_+) (@_@)": "amazed_3",
    "(*^^)v": "laughing_cheerful_1",
    "(^_^)v": "laughing_cheerful_2",
    "((d[-_-]b))": "headphones_listening_to_music",
    '(-"-)': "worried_1",
    "(ーー;)": "worried_2",
    "(^0_0^)": "eyeglasses",
    "(＾ｖ＾)": "happy_1",
    "(＾ｕ＾)": "happy_2",
    "(^)o(^)": "happy_3",
    "(^O^)": "happy_4",
    "(^o^)": "happy_5",
    ")^o^(": "happy_6",
    "o_O": "surprised_1",
    "o_0": "surprised_2",
    "o.O": "surprised_6",
    "O_O": "surprised_4",
    "(o.o)": "surprised_5",
    "(*￣m￣)": "dissatisfied",
    "(‘A`)": "snubbed_or_deflated",
    "uwu": "otaku_tenderness_1",
    "UwU": "otaku_tenderness_2",
    "owo": "otaku_tenderness_3",
    "OwO": "otaku_tenderness_4",
    ">w<": "otaku_tenderness_5",
    ">W<": "otaku_tenderness_6",
    "^w^": "otaku_tenderness_7",
    "^W^": "otaku_tenderness_8",
}


def demoticonize(string, delimiters=(" _", "_ ")):
    """Replace emoticons with their corresponding text in the dictionary EMOTICONS"""
    for emoticon, text in EMOTICONS.items():
        string = string.replace(emoticon, f"{delimiters[0]}{text}{delimiters[1]}")
    return string


def emoticonize(string, delimiters=(" _", "_ ")):
    """Replace text in the dictionary EMOTICONS with their corresponding emoticons"""
    for emoticon, text in EMOTICONS.items():
        string = string.replace(f"{delimiters[0]}{text}{delimiters[1]}", emoticon)
    return string



"""
Inclusive words data dictonary
"""

INCLUSIVE_WORDS = {
    "nosotres": "nosotros",
    "nosotrxs": "nosotros",
    "nosotr@s": "nosotros",
    "todes": "todos",
    "todxs": "todos",
    "tod@s": "todos",
    "amiges": "amigos",
    "amigxs": "amigos",
    "amig@s": "amigos",
    "compañeres": "compañeros",
    "compañerxs": "compañeros",
    "compañer@s": "compañeros",
    "chiques": "chicos",
    "chicxs": "chicos",
    "chic@s": "chicos",
    "niñes": "niños",
    "niñxs": "niños",
    "niñ@s": "niños",
    "elle": "él",
    "ellxs": "ellos",
    "ell@s": "ellos",
    "les": "los",
}

def normalize_inclusive_language(text: str) -> str:
    """
    Normalize inclusive language
    """
    for key, value in INCLUSIVE_WORDS.items():
        if key in text:
            text = text.replace(key, value)
    return text



import re

import numpy as np
from math import ceil
from unidecode import unidecode
import json
import os

from .abstract import DataAugmentationAbstract
import random


class Spelling(DataAugmentationAbstract):
    """
    Class for data augmentation with spelling.
    """

    def __init__(
        self, method, stopwords="default", aug_percent=0.1, tokenizer="default"
    ):
        """
        Initialize the class.
        """
        super().__init__(method)

        available_methods = [
            "keyboard",
            "ocr",
            "random",
            "grapheme_spelling",
            "word_spelling",
            "remove_punctuation",
            "remove_accents",
            "remove_spaces",
            "lowercase",
            "uppercase",
            "randomcase",
            "all",
        ]

        if self.method not in available_methods:
            str_methods = ", ".join(available_methods)
            raise ValueError(
                f"Method not available. The method must be {str_methods}."
            )

        self.aug_percent = aug_percent
        self.tokenizer = tokenizer
        self.stopwords = stopwords

        if self.tokenizer == "default":
            self._load_default_tokenizer_()
        if self.stopwords == "default":
            self._load_default_stopwords_()

    def set_aug_percent(self, aug_percent):
        """
        Set the augmentation percentage.
        """
        if not isinstance(aug_percent, float):
            raise ValueError("The aug_percent must be a float.")
        self.aug_percent = aug_percent

    def set_tokenizer(self, tokenizer):
        """
        Set the tokenizer.
        """
        self.tokenizer = tokenizer

    def _text_augment_(self, text, num_samples=1):
        """
        Augment a single text.
        """
        if self.method == "keyboard":
            return self._keyboard_augment_(text, num_samples)
        elif self.method == "ocr":
            return self._ocr_augment_(text, num_samples)
        elif self.method == "random":
            return self._random_augment_(text, num_samples)
        elif self.method == "grapheme_spelling":
            return self._grapheme_spelling_augment_(text, num_samples)
        elif self.method == "word_spelling":
            return self._word_spelling_augmentation_(text, num_samples)
        elif self.method == "remove_punctuation":
            return self._remove_punctuation_augment_(text, num_samples)
        elif self.method == "remove_spaces":
            return self._remove_spaces_augment_(text, num_samples)
        elif self.method == "remove_accents":
            return self._remove_accents_augmentation_(text, num_samples)
        elif self.method == "lowercase":
            return self._lowercase_augmentation_(text, num_samples)
        elif self.method == "uppercase":
            return self._uppercase_augmentation_(text, num_samples)
        elif self.method == "randomcase":
            return self._random_case_augmentation_(text, num_samples)
        elif self.method == "all":
            return self._all_augment_(text, num_samples)
        else:
            raise ValueError("Invalid method")

    def _set_keyboard_augment_dict_(self):
        """Create keyboard augment dict for qwerty keyboard"""
        self.keyboard_augment_dict = {
            "q": ["w", "a", "s"],
            "w": ["q", "e", "a", "s", "d"],
            "e": ["w", "s", "d", "f", "r"],
            "r": ["e", "d", "f", "g", "t"],
            "t": ["r", "f", "g", "h", "y"],
            "y": ["t", "g", "h", "j", "u"],
            "u": ["y", "h", "j", "k", "i"],
            "i": ["u", "j", "k", "l", "o"],
            "o": ["i", "k", "l", "p"],
            "p": ["o", "l", "ñ"],
            "a": ["q", "w", "s", "z", "x"],
            "s": ["a", "q", "w", "e", "d", "z", "x"],
            "d": ["s", "w", "e", "r", "f", "x", "c"],
            "f": ["d", "e", "r", "t", "g", "c", "v"],
            "g": ["f", "r", "t", "y", "h", "v", "b"],
            "h": ["g", "t", "y", "u", "j", "b", "n"],
            "j": ["h", "y", "u", "i", "k", "n", "m"],
            "k": ["j", "u", "i", "o", "l", "m"],
            "l": ["k", "i", "o", "p", "ñ"],
            "ñ": ["l", "o", "p"],
            "z": ["a", "s", "x"],
            "x": ["z", "s", "d", "c"],
            "c": ["x", "d", "f", "v"],
            "v": ["c", "f", "g", "b"],
            "b": ["v", "g", "h", "n"],
            "n": ["b", "h", "j", "m"],
            "m": ["n", "j", "k"],
        }

    def _keyboard_augment_(self, text, num_samples):
        """
        Increase textual data by modifying characters randomly according to the proximity of the keystrokes
        """
        # If self.keyboard_augment_dict is not defined, define it
        if not hasattr(self, "keyboard_augment_dict"):
            self._set_keyboard_augment_dict_()

        # List to save the augmented texts
        output_texts = []

        # Iterate over the number of samples
        for i in range(num_samples):
            # Get the number of characters to augment
            num_aug = int(len(text) * self.aug_percent)

            # Get the indices of the characters to augment
            aug_indices = np.random.choice(
                range(len(text)), num_aug, replace=False
            )
            # Get the characters to augment
            aug_chars = [text[i] for i in aug_indices]
            # Get the augmented characters
            aug_chars = [
                self.keyboard_augment_dict[c][
                    np.random.randint(0, len(self.keyboard_augment_dict[c]))
                ]
                if c in self.keyboard_augment_dict
                else c
                for c in aug_chars
            ]
            # Augment the text
            output_text = list(text)
            for j in range(num_aug):
                output_text[aug_indices[j]] = aug_chars[j]
            output_text = "".join(output_text)
            # Append the augmented text to the list
            output_texts.append(output_text)

        return output_texts

    def _set_ocr_augment_dict_(self):
        self.ocr_augment_dict = {
            "0": ["D", "0", "o", "8"],
            "1": ["I", "l", "i", "L"],
            "2": ["Z", "z"],
            "3": ["E", "e"],
            "4": ["A", "a"],
            "5": ["S", "s"],
            "6": ["G", "g"],
            "7": ["T", "t"],
            "8": ["B", "s", "S"],
            "9": ["g", "q"],
            "A": ["4"],
            "B": ["8", "E"],
            "C": ["G"],
            "D": ["0", "O"],
            "E": ["3"],
            "F": ["T"],
            "G": ["6", "C"],
            "H": ["4"],
            "I": ["1", "l", "i", "L"],
            "J": ["L"],
            "K": ["X"],
            "L": ["1", "i", "I"],
            "M": ["W"],
            "N": ["H"],
            "Ñ": ["N", "n"],
            "O": ["0", "Q"],
            "P": ["R"],
            "Q": ["O", "9"],
            "R": ["P"],
            "S": ["5", "8", "B"],
            "T": ["7"],
            "U": ["V", "v"],
            "V": ["U"],
            "X": ["K"],
            "Y": ["V"],
            "Z": ["2", "Z"],
            "b": ["8", "E"],
            "c": ["G"],
            "d": ["0", "O"],
            "e": ["3"],
            "f": ["T"],
            "g": ["6", "C", "9", "q"],
            "h": ["4"],
            "i": ["1", "l", "I"],
            "j": ["L"],
            "k": ["X"],
            "l": ["1", "i", "I"],
            "m": ["W"],
            "ñ": ["N", "n"],
            "o": ["0", "Q"],
            "p": ["R"],
            "q": ["O", "9", "g"],
            "r": ["P"],
            "s": ["5", "8", "B"],
            "t": ["7"],
            "u": ["V"],
            "v": ["U"],
            "w": ["M", "V"],
            "x": ["K"],
            "y": ["V"],
            "z": ["2", "Z"],
        }

    def _ocr_augment_(self, text, num_samples):
        """
        Increase textual data by modifying characters randomly according to the common OCR errors for Spanish
        """
        # If self.keyboard_augment_dict is not defined, define it
        if not hasattr(self, "ocr_augment_dict"):
            self._set_ocr_augment_dict_()

        # List to save the augmented texts
        output_texts = []

        # Iterate over the number of samples
        for i in range(num_samples):
            # Get the number of characters to augment
            num_aug = int(len(text) * self.aug_percent)
            # Get the indices of the characters to augment
            aug_indices = np.random.choice(
                range(len(text)), num_aug, replace=False
            )

            # Get the characters to augment
            aug_chars = [text[i] for i in aug_indices]

            # Get the augmented characters
            aug_chars = [
                self.ocr_augment_dict[c][
                    np.random.randint(0, len(self.ocr_augment_dict[c]))
                ]
                if c in self.ocr_augment_dict
                else c
                for c in aug_chars
            ]
            aug_text = list(text)
            for i, c in zip(aug_indices, aug_chars):
                aug_text[i] = c
            aug_text = "".join(aug_text)
            # Add the augmented text to the list
            output_texts.append(aug_text)
        return output_texts

    def _set_alphabet_(self):
        """
        Set the alphabet of the dataset
        """
        alphabet = []
        alphabet.extend([chr(i) for i in range(ord("A"), ord("Z") + 1)])
        alphabet.extend([chr(i) for i in range(ord("a"), ord("z") + 1)])
        alphabet.extend([chr(i) for i in range(ord("0"), ord("9") + 1)])
        alphabet.extend(
            [
                "á",
                "é",
                "í",
                "ó",
                "ú",
                "ñ",
                "ü",
                "Á",
                "É",
                "Í",
                "Ó",
                "Ú",
                "Ñ",
                "Ü",
            ]
        )
        self.alphabet = alphabet

    def _random_augment_(self, text, num_samples):
        """
        Increase textual data by modifying characters randomly
        """
        if not hasattr(self, "alphabet"):
            self._set_alphabet_()

        # List to save the augmented texts
        output_texts = []

        # Iterate over the number of samples
        for i in range(num_samples):
            # Get the number of characters to augment
            num_aug = int(len(text) * self.aug_percent)
            # Get the indices of the characters to augment
            aug_indices = np.random.choice(
                range(len(text)), num_aug, replace=False
            )
            # Get the characters to augment
            aug_chars = [text[i] for i in aug_indices]
            # Get the augmented characters if they are in the alphabet
            aug_chars = [
                np.random.choice(self.alphabet) if c in self.alphabet else c
                for c in aug_chars
            ]
            # Augment the text
            output_text = list(text)
            for j in range(num_aug):
                output_text[aug_indices[j]] = aug_chars[j]

            output_text = "".join(output_text)
            # Append the augmented text to the list
            output_texts.append(output_text)
        return output_texts

    def _set_grapheme_spelling_dict_(self):
        """
        Set the grapheme_spelling dictionary
        """
        grapheme_spelling_dict = {
            "mb": "nv",
            "nv": "mb",
            "m": "n",
            "n": "m",
            "v": "b",
            "b": "v",
            "ll": "y汉",  # then replace "汉" to ""
            "y": "漢",  # then replace "漢" to "ll",
            "h": " ",
            "qu": "k汉",  # then replace "汉" to ""
            "g": "j",
            "j": "g",
            "z": "s",
            "ci": "si",
        }

        self.grapheme_spelling_dict = grapheme_spelling_dict

    def __find_substring_indexes__(self, string, substring):
        """Find all indexes of a substring in a string.

        Args:
            string (str): string to search in.
            substring (str): substring to search for.

        Returns:
            list: list of tuples with the start and end indexes of the substring.
        """
        pattern = re.compile(re.escape(substring))
        matches = pattern.finditer(string)
        return [(match.start(), match.end() - 1) for match in matches]

    def _grapheme_spelling_augment_(self, text, num_samples):
        """
        Increase textual data by modifying characters randomly according to the common grapheme_spellings for Spanish

        """
        # If self.grapheme_spelling_dict is not defined, define it
        if not hasattr(self, "grapheme_spelling_dict"):
            self._set_grapheme_spelling_dict_()
        # List to save the augmented texts
        output_texts = []

        # Count the times that are self.keyboard_augment_dict keys in the text with self.__find_substring_indexes__(string, substring)
        aparitions = [
            self.__find_substring_indexes__(text, key)
            for key in self.grapheme_spelling_dict.keys()
        ]
        # Join all the sublists in a single list
        # Get the number of characters to augment
        aparitions = sum(aparitions, [])
        num_aug = int(len(aparitions) * self.aug_percent)
        # Iterate over the number of samples
        for i in range(num_samples):
            new_text = text
            # Create a list with num_aug elements in aparitions without repetition
            elements = random.sample(aparitions, num_aug)
            for e in elements:
                start = e[0]
                end = e[1] + 1
                substring = new_text[start:end]
                replacement = self.grapheme_spelling_dict[substring]
                # Select a random element from the list of replacements
                # replacement = random.choice(replacement)
                start_str = new_text[:start]
                end_str = new_text[end:] if end < len(new_text) else ""
                new_text = start_str + replacement + end_str

            # Replace auxiliar characters
            new_text = new_text.replace("汉", "")
            new_text = new_text.replace("漢", "ll")
            # Append the augmented text to the list
            output_texts.append(new_text)
        return output_texts

    def _remove_punctuation_augment_(self, text, num_samples):
        """Remove punctuation in the text according to the aug_percent

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # List to save the augmented texts
        output_texts = []
        # List with the punctuation to eliminate
        punctuation = [".", ",", "¡", "¿"]
        num_text_punctuation = [c for c in text if c in punctuation]
        num_aug = len(num_text_punctuation) * self.aug_percent
        # Round by the upper integer
        num_aug = ceil(num_aug)
        for i in range(num_samples):
            new_text = text
            # Get the indices of the punctuation to eliminate
            indices = [i for i, c in enumerate(new_text) if c in punctuation]
            # Create a list with num_aug elements in indices without repetition
            elements = random.sample(indices, num_aug)
            for e in elements:
                new_text = new_text[:e] + new_text[e + 1 :]
            # Append the augmented text to the list
            output_texts.append(new_text)
        return output_texts

    def _remove_spaces_augment_(self, text, num_samples):
        """Remove spaces in the text according to the aug_percent

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # List to save the augmented texts
        output_texts = []
        # List with the punctuation to eliminate
        num_text_spaces = [c for c in text if c == " "]
        num_aug = len(num_text_spaces) * self.aug_percent
        # Round by the upper integer
        num_aug = ceil(num_aug)
        for i in range(num_samples):
            new_text = text
            # Get the indices of the punctuation to eliminate
            indices = [i for i, c in enumerate(new_text) if c == " "]
            # Create a list with num_aug elements in indices without repetition
            elements = random.sample(indices, num_aug)
            for e in elements:
                new_text = new_text[:e] + new_text[e + 1:]
            # Append the augmented text to the list
            output_texts.append(new_text)
        return output_texts

    def _remove_accents_augmentation_(self, text, num_samples):
        """Remove accents in the text according to the aug_percent

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # List to save the augmented texts
        output_texts = []
        # List with the punctuation to eliminate
        accents = ["á", "é", "í", "ó", "ú", "Á", "É", "Í", "Ó", "Ú", "ü", "Ü"]
        num_text_accents = [c for c in text if c in accents]
        num_aug = len(num_text_accents) * self.aug_percent
        # Round by the upper integer
        num_aug = ceil(num_aug)
        for i in range(num_samples):
            new_text = text
            # Get the indices of the punctuation to eliminate
            indices = [i for i, c in enumerate(new_text) if c in accents]
            # Create a list with num_aug elements in indices without repetition
            elements = random.sample(indices, num_aug)
            # Replace every element with accent with the same element without accent
            for e in elements:
                new_e = unidecode(new_text[e])
                new_text = new_text[:e] + new_e + new_text[e + 1 :]

            # Append the augmented text to the list
            output_texts.append(new_text)
        return output_texts

    def _lowercase_augmentation_(self, text, num_samples):
        """Randomly chooses words containing at least one capital letter
        and converts them to lowercase according to the aug_percent.

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # List to save the augmented texts
        output_texts = []
        uppercase_words = []
        word = ""
        start = 0
        # Add to uppercase_words the words (separated by spaces) with at least one letter uppercase ({'word': 'word', 'start': first_int, 'end': last_int}})
        for index, character in enumerate(text):
            if character != " ":
                word += character
            else:
                end = index
                if any(char.isupper() for char in word):
                    uppercase_words.append(
                        {
                            "word": word,
                            "start": index - len(word),
                            "end": index,
                        }
                    )
                start = index + 1
                word = ""

        # Num of words to lowercase
        num_aug = ceil(len(uppercase_words) * self.aug_percent)
        for i in range(num_samples):
            # Copy the original text
            new_text = text
            # Choose a random subset of words to lowercase
            selected_words = random.sample(uppercase_words, num_aug)
            # Lowercase the selected words in the new text
            for word in selected_words:
                new_word = word["word"].lower()
                new_text = (
                    new_text[: word["start"]]
                    + new_word
                    + new_text[word["end"] :]
                )
            output_texts.append(new_text)
        return output_texts

    def _uppercase_augmentation_(self, text, num_samples):
        """Randomly chooses words containing at least one lower letter
        and converts them to lowercase according to the aug_percent.

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # List to save the augmented texts
        output_texts = []
        lowercase_worsd = []
        word = ""
        start = 0
        # Add to lowercase_worsd the words (separated by spaces) with at least one letter lowercase ({'word': 'word', 'start': first_int, 'end': last_int}})
        for index, character in enumerate(text):
            if character != " ":
                word += character
            else:
                end = index
                if any(char.islower() for char in word):
                    lowercase_worsd.append(
                        {
                            "word": word,
                            "start": index - len(word),
                            "end": index,
                        }
                    )
                start = index + 1
                word = ""

        # Num of words to lowercase
        num_aug = ceil(len(lowercase_worsd) * self.aug_percent)
        for i in range(num_samples):
            # Copy the original text
            new_text = text
            # Choose a random subset of words to lowercase
            selected_words = random.sample(lowercase_worsd, num_aug)
            # Lowercase the selected words in the new text
            for word in selected_words:
                new_word = word["word"].upper()
                new_text = (
                    new_text[: word["start"]]
                    + new_word
                    + new_text[word["end"] :]
                )
            output_texts.append(new_text)
        return output_texts

    def _random_case_augmentation_(self, text, num_samples):
        """Randomly lowercase or uppercase words in the text according to the aug_percent

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # List to save the augmented texts
        output_texts = []
        half_aug_percent = self.aug_percent / 2

        for i in range(num_samples):
            # Copy the original text
            new_text = text
            # Lowercase the text
            new_text = new_text.lower()
            # Uppercase the text
            new_text = self._uppercase_augmentation_(new_text, 1)[0]
            # Lowercase the text
            new_text = self._lowercase_augmentation_(new_text, 1)[0]
            # Append the augmented text to the list
            output_texts.append(new_text)
        return output_texts

    def _load_mispelled_words_dictionary_(self):
        """
        Set the dictionary of misspelled words
        """
        # Load the dictionary
        json_path = os.path.join(
            os.path.dirname(__file__), "data", "misspelled_words.json"
        )
        with open(json_path, "r") as f:
            self.misspelled_dict = json.load(f)

    def _word_spelling_augmentation_(self, text, num_samples):
        """Randomly chooses words from a dictionary and replaces them with a random misspelled word

        Args:
            text (str): text to augment
            num_samples (int): number of samples to generate
        """
        # If self.misspelled_dict is not defined, define it
        if not hasattr(self, "misspelled_dict"):
            self._load_mispelled_words_dictionary_()
        output_texts = []
        # Split the text in words
        words = text.split(" ")
        # Check words in the dictionary
        words = [word for word in words if word in self.misspelled_dict]
        # Num of words to replace
        num_aug = ceil(len(words) * self.aug_percent)
        for i in range(num_samples):
            # Copy the original text
            new_text = text
            # Choose a random subset of words to replace
            selected_words = random.sample(words, num_aug)
            # Replace the selected words in the new text
            for word in selected_words:
                # Select a random misspelled word
                new_word = random.choice(self.misspelled_dict[word])
                new_text = new_text.replace(word, new_word)
            output_texts.append(new_text)
        return output_texts

    def _all_augment_(self, text, num_samples):
        """
        Increase textual data by modifying characters randomly according to the common grapheme_spellings for Spanish
        """
        aug_percent = self.aug_percent
        self.aug_percent = aug_percent / 7
        output_texts = []
        for i in range(num_samples):
            text = self._word_spelling_augmentation_(text, 1)
            text = self._grapheme_spelling_augment_(text[0], 1)
            text = self._keyboard_augment_(text[0], 1)
            text = self._ocr_augment_(text[0], 1)
            text = self._random_augment_(text[0], 1)
            text = self._remove_punctuation_augment_(text[0], 1)
            text = self._remove_spaces_augment_(text[0], 1)
            text = self._remove_accents_augmentation_(text[0], 1)
            text = self._random_case_augmentation_(text[0], 1)
            output_texts.append(text[0])
        self.aug_percent = aug_percent
        return output_texts



import re
import string

import emoji
from nltk.stem.snowball import SnowballStemmer

from spanish_nlp.utils.emo_unicode import demoticonize, emoticonize
from spanish_nlp.utils.inclusive_words import normalize_inclusive_language


class SpanishPreprocess:
    def __init__(
        self,
        lower=True,
        remove_url=True,
        remove_hashtags=False,
        split_hashtags=True,
        normalize_breaklines=True,
        remove_emoticons=True,
        remove_emojis=True,
        convert_emoticons=False,
        convert_emojis=False,
        normalize_inclusive_language=False,
        reduce_spam=True,
        remove_reduplications=True,
        remove_vowels_accents=True,
        remove_multiple_spaces=True,
        remove_punctuation=True,
        remove_unprintable=True,
        remove_numbers=True,
        remove_stopwords=False,
        stopwords_list=None,
        lemmatize=False,
        stem=False,
        remove_html_tags=True,
    ):
        """A class for preprocessing Spanish text for NLP tasks.

        Args:
            lower (bool, optional): convert text to lowercase. Defaults to True.
            remove_url (bool, optional): remove urls from text. Defaults to True.
            remove_hashtags (bool, optional): remove hashtags from text. Defaults to False.
            split_hashtags (bool, optional): split hashtags from text. Defaults to True.
            normalize_breaklines (bool, optional): remove multiple breaklines. Defaults to True.
            remove_emoticons (bool, optional): remove emoticons from text. Defaults to True.
            remove_emojis (bool, optional): remove emojis from text. Defaults to True.
            convert_emoticons (bool, optional): codify emoticons to text. Defaults to False.
            convert_emojis (bool, optional): codify emojis to text. Defaults to False.
            normalize_inclusive_language (bool, optional): normalize inclusive language. Defaults to False.
            reduce_spam (bool, optional): remove repeated words. Defaults to True.
            remove_vowels_accents (bool, optional): remove vowels with accents. Defaults to True.
            remove_multiple_spaces (bool, optional): remove multiple spaces. Defaults to True.
            remove_punctuation (bool, optional): remove punctuation. Defaults to True.
            remove_unprintable (bool, optional): remove unprintable characters. Defaults to True.
            remove_numbers (bool, optional): remove numbers. Defaults to True.
            remove_stopwords (bool, optional): remove stopwords. Defaults to False.
            stopwords_list (str, list, optional): stopwords name ('default', 'extended', 'nltk', 'spacy') or list of stopwords. Defaults to None.
            lemmatize (bool, optional): lemmatize text. Defaults to False.
            stem (bool, optional): stem text. Defaults to False.
            remove_html_tags (bool, optional): remove html tags. Defaults to True.
        """
        self.lower = lower
        self.remove_url = remove_url
        self.remove_hashtags = remove_hashtags
        self.split_hashtags = split_hashtags
        self.normalize_breaklines = normalize_breaklines
        self.remove_emojis = remove_emojis
        self.remove_emoticons = remove_emoticons
        self.convert_emoticons = convert_emoticons
        self.convert_emojis = convert_emojis
        self.normalize_inclusive_language = normalize_inclusive_language
        self.reduce_spam = reduce_spam
        self.remove_reduplications = remove_reduplications
        self.remove_vowels_accents = remove_vowels_accents
        self.remove_multiple_spaces = remove_multiple_spaces
        self.remove_punctuation = remove_punctuation
        self.remove_numbers = remove_numbers
        self.remove_stopwords = remove_stopwords
        self._prepare_stopwords_(stopwords_list)
        self.remove_unprintable = remove_unprintable
        self.stem = stem
        self.lemmatize = lemmatize
        self.remove_html_tags = remove_html_tags
        self.normalize_punctuation_spelling = True

        self._check_errors_()
        self._prepare_lemmatize_()

    def _check_errors_(self):
        if self.lemmatize and self.stem:
            raise ValueError("Lemmatize and Stem are exclusive. Choose one.")

        if self.remove_emojis and self.convert_emojis:
            raise ValueError(
                "Remove emojis and convert emojis are exclusive. Choose one."
            )
        if self.remove_emoticons and self.convert_emoticons:
            raise ValueError(
                "Remove emoticons and convert emoticons are exclusive. Choose one."
            )
        if self.split_hashtags and self.remove_hashtags:
            raise ValueError(
                "Split hashtags and remove hashtags are exclusive. Choose one."
            )
        if self.remove_stopwords and self.stopwords_list is None:
            raise ValueError(
                "If remove stopwords is True, you must provide a type of stopwords list ('default', 'extended', 'nltk', 'spacy') or a list of stopwords."
            )

    def _prepare_stopwords_(self, type="default"):
        if type == "default":
            from spanish_nlp.utils.stopwords import default_stopwords

            self.stopwords_list = default_stopwords

        elif type == "extended":
            from spanish_nlp.utils.stopwords import extended_stopwords

            self.stopwords_list = extended_stopwords

        elif type == "nltk":
            import nltk

            nltk.download("stopwords")
            from nltk.corpus import stopwords

            self.stopwords_list = stopwords.words("spanish")
        elif type == "spacy":
            import es_core_news_sm

            nlp = es_core_news_sm.load(
                disable=["ner", "parser", "tagger", "textcat", "vectors"]
            )
            self.stopwords_list = nlp.Defaults.stop_words
            del nlp

        elif type is None:
            self.stopwords_list = None

        else:
            if isinstance(type, list):
                self.stopwords_list = type
            else:
                raise ValueError(
                    "Stopwords must be a list or one of the following: 'default', 'extended', 'nltk', 'spacy'"
                )

    def _prepare_lemmatize_(self, force=False):
        if self.lemmatize or force:
            import es_core_news_sm

            self.nlp_spacy = es_core_news_sm.load(
                disable=["ner", "parser", "tagger", "textcat", "vectors"]
            )

    def _debug_method_(self, text, method_name):
        print(f"Method: {method_name}")
        print("---" * 30)
        print(f"Text: {text}")
        print("===" * 40)

    def _lower_(self, text):
        return text.lower()

    def _remove_url_(self, text):
        """Remove urls from text. By example:
        "Este es un texto con una url: https://www.google.com" -> "Este es un texto con una url: "
        "Una URL como http://page.com/page/test?param=1&param2=2 tiene parámetros" -> "Una URL como tiene parámetros"
        """
        url_pattern = re.compile(r"https?://\S+")
        text = url_pattern.sub("", text).replace("  ", " ")
        return text

    def _remove_hashtags_(self, text):
        """Remove hashtags from text. By example:
        "Este es un texto con un hashtag: #hashtag" -> "Este es un texto con un hashtag:"
        "Tengo un #hashtag1 #HashTag2 y #hasTag3" -> "Tengo un y"
        """
        return re.sub(r"#\w+", "", text).strip().replace("  ", " ")

    def _split_hashtags_(self, text):
        """Split hashtags from text.
        Example:
            "este es #unEjemplo de #TextosConHashtag #SiSeñor #yey." -> "este es un Ejemplo de Textos Con Hashtag Si Señor yey."
        """
        # Find all hashtags
        hashtags = re.findall(r"(?<=\s)#(\w+)", text)
        # Delete hashtag with numbers
        hashtags = [ht for ht in hashtags if not re.search(r"\d", ht)]
        # Split all hashtags and replace them in the text
        pattern = re.compile(
            r"[A-ZÑÁÉIÓÚ]*[a-zñáéíóúü0-9]+|\d+|[A-ZÑÁÉIÓÚ]+(?![a-zñáéíóúü])"
        )
        for ht in hashtags:
            words = " ".join(pattern.findall(ht)).strip()
            text = text.replace(f"#{ht}", f"{words}")
        return text.replace("  ", " ").strip()

    def _normalize_breaklines_(self, text):
        """Convert multiple breaklines to one breakline"""
        text = text.replace("\r", "\n")
        # text = re.sub(r"(\n){2,}", r"\n", text)
        # Can there are 0 or more spaces between breaklines
        text = re.sub(r"(\n\s*)+", r"\n", text)
        return text.strip()

    def _emoticons_to_text_(self, text):
        pp_text = demoticonize(text, delimiters=(" __", "__ "))
        return self._normalize_punctuation_spelling_(pp_text)

    def _emojis_to_text_(self, text):
        pp_text = emoji.demojize(text, delimiters=(" __", "__ ")).replace("  ", " ")
        return self._normalize_punctuation_spelling_(pp_text)

    def _text_to_emojis_(self, text):
        pp_text = emoji.emojize(text, delimiters=("__", "__"))
        return self._normalize_punctuation_spelling_(pp_text)

    def _text_to_emoticons_(self, text):
        pp_text = emoticonize(text, delimiters=("__", "__"))
        return pp_text

    def _normalize_inclusive_language_(self, text):
        """Replace inclusive language with a dictionary of some words"""
        return normalize_inclusive_language(text)

    def _reduce_spam_(self, text):
        """Reduce a expression if it is repeated more than 3 times and convert it to two expressions.
        Example: "hola hola hola hola hola hola" -> "hola hola hola"
        """
        text = re.sub(r"(\w+\s)\1+", r"\1\1", text)
        text = re.sub(r"(\b(\w+\s){3})\1+", r"\1\1", text)
        return text

    def _remove_reduplications_(self, text):
        """Use a regular expression to find a sequence of non-digit characters
        that are repeated at the end of the word, and replace it with just
        one instance of the character

        Examples: "holaaa cómo estás?" -> "hola cómo estás?"
                  "no te creoooo naaada" -> "no te creo naaada"
        """
        pattern = r"([aeiou])\1+"
        return re.sub(pattern, r"\1", text)

    def _remove_vowels_accents_(self, text):
        """Convert vowels with accents from text (lowercase or uppercase)"""
        text = re.sub(r"[áàäâ]", "a", text)
        text = re.sub(r"[éèëê]", "e", text)
        text = re.sub(r"[íìïî]", "i", text)
        text = re.sub(r"[óòöô]", "o", text)
        text = re.sub(r"[úùüû]", "u", text)
        text = re.sub(r"[ÁÀÄÂ]", "A", text)
        text = re.sub(r"[ÉÈËÊ]", "E", text)
        text = re.sub(r"[ÍÌÏÎ]", "I", text)
        text = re.sub(r"[ÓÒÖÔ]", "O", text)
        text = re.sub(r"[ÚÙÜÛ]", "U", text)
        return text

    def _remove_punctuation_(self, text):
        pattern = re.compile(r"[^\w\sáéíóúüñÁÉÍÓÚÜÑ]")
        t = pattern.sub(r" ", text)
        return re.sub(" +", " ", t)

    def _remove_unprintable_(self, text):
        printable = set(string.printable + "ñáéíóúü" + "ÑÁÉÍÓÚÜ")
        text = "".join(filter(lambda x: x in printable, text))
        return text

    def _remove_numbers_(self, text):
        """Remove numbers from text"""
        return re.sub(r"\d+", "", text)

    def _convert_numbers_(self, text, delimiters=(" __", "__ ")):
        return re.sub(r"[0-9]", delimiters[0] + "numero" + delimiters[1], text)

    def _remove_stopwords_(self, text):
        return " ".join(
            [
                word
                for word in str(text).split()
                if word.lower() not in self.stopwords_list
            ]
        )

    def _stem_(self, text, stemmer=SnowballStemmer("spanish")):
        """TODO: add another stemmers"""
        return " ".join([stemmer.stem(word) for word in text.split()])

    def _lemmatize_(self, text, lemmatizer="es_core_news_sm"):
        """Lemmatize text using es_core_news_sm from spacy by default"""
        doc = self.nlp_spacy(text)
        return " ".join([token.lemma_ for token in doc])

    def _remove_multiples_spaces_(self, text):
        return re.sub(" +", " ", text).strip()

    def _normalize_punctuation_spelling_(self, text):
        """Remove all wrong spaces with punctuation"""
        # Remove spaces before punctuation
        text = re.sub(r" +([\.\,\!\?\)\]\}\>\:\#}])", r"\1", text)
        # Remove spaces after punctuation
        text = re.sub(r"([\¡\¿\(\[\{\<])\: +", r"\1", text)
        # Add space after , and . if it is not a number or an url and it does not have a space
        text = re.sub(r"([\.\,])([^\s\d])", r"\1 \2", text)
        # Remove duplicated spaces
        return self._remove_multiples_spaces_(text).strip()

    def _remove_html_tags_(self, text):
        """Remove html tags from a string"""
        import re

        clean = re.compile("<.*?>")
        return re.sub(clean, "", text)

    def transform(self, text, debug=False):
        if self.split_hashtags:
            text = self._split_hashtags_(text)
            self._debug_method_(text, "split_hashtags") if debug else None

        if self.lower:
            text = self._lower_(text)
            self._debug_method_(text, "lower") if debug else None

        if self.remove_url:
            text = self._remove_url_(text)
            self._debug_method_(text, "remove_url") if debug else None

        if self.remove_html_tags:
            text = self._remove_html_tags_(text)
            self._debug_method_(text, "remove_html_tags") if debug else None

        if self.remove_numbers:
            text = self._remove_numbers_(text)
            self._debug_method_(text, "remove_numbers") if debug else None

        if self.remove_hashtags:
            text = self._remove_hashtags_(text)
            self._debug_method_(text, "remove_hashtags") if debug else None

        if self.stem:
            text = self._stem_(text)
            self._debug_method_(text, "stem") if debug else None

        if self.lemmatize:
            text = self._lemmatize_(text)
            self._debug_method_(text, "lemmatize") if debug else None

        if self.convert_emojis or not self.remove_emojis:
            text = self._emojis_to_text_(text)
            self._debug_method_(text, "emojis_to_text") if debug else None

        if self.convert_emoticons or not self.remove_emoticons:
            text = self._emoticons_to_text_(text)
            self._debug_method_(text, "emoticons_to_text") if debug else None

        if self.normalize_inclusive_language:
            text = self._normalize_inclusive_language_(text)
            self._debug_method_(text, "normalize_inclusive_language") if debug else None

        if self.remove_vowels_accents:
            text = self._remove_vowels_accents_(text)
            self._debug_method_(text, "remove_vowels_accents") if debug else None

        if self.remove_punctuation:
            text = self._remove_punctuation_(text)
            self._debug_method_(text, "remove_punctuation") if debug else None

        if self.remove_unprintable:
            text = self._remove_unprintable_(text)
            self._debug_method_(text, "remove_unprintable") if debug else None

        if self.stem:
            text = self._stem_(text)
            self._debug_method_(text, "stem") if debug else None

        if not self.remove_emojis:
            text = self._text_to_emojis_(text)
            self._debug_method_(
                text, "text_to_emojis (not delete emojis)"
            ) if debug else None

        if not self.remove_emoticons:
            text = self._text_to_emoticons_(text)
            self._debug_method_(
                text, "text_to_emoticons (not delete emoticons)"
            ) if debug else None

        if self.remove_stopwords:
            text = self._remove_stopwords_(text)
            self._debug_method_(text, "remove_stopwords") if debug else None

        if self.remove_multiple_spaces:
            text = self._remove_multiples_spaces_(text).strip()
            self._debug_method_(text, "remove_multiples_spaces") if debug else None

        if self.normalize_breaklines:
            text = self._normalize_breaklines_(text)
            self._debug_method_(text, "normalize_breaklines") if debug else None

        if self.normalize_punctuation_spelling:
            text = self._normalize_punctuation_spelling_(text)
            self._debug_method_(
                text, "normalize_punctuation_spelling"
            ) if debug else None

        if self.reduce_spam:
            text = self._reduce_spam_(text)
            self._debug_method_(text, "reduce_spam") if debug else None

        if self.remove_reduplications:
            text = self._remove_reduplications_(text)
            self._debug_method_(text, "remove_reduplications") if debug else None

        return text



import torch
from transformers import pipeline


class SpanishClassifier:
    def __init__(self, model_name=None, device=None):
        self.model_name = model_name
        self.device = device
        if self.device is None:
            self._set_default_device_()
        self.model = NotImplemented
        self.labels = NotImplemented
        self.last_prediction = NotImplemented
        self.multiclass = NotImplemented
        if self.model_name is not None:
            self._configure_model_()

    def _set_default_device_(self):
        self.device = 0 if torch.cuda.is_available() else -1

    def set_model(self, model_name):
        self.model_name = model_name

    def _configure_model_(self):
        if self.model_name == "hate_speech":
            self.load_hate_speech()
        elif self.model_name == "incivility":
            self.load_incivility()
        elif self.model_name == "toxic_speech":
            self.load_toxic_speach()
        elif self.model_name == "sentiment_analysis":
            self.load_sentiment_analysis()
        elif self.model_name == "emotion_analysis":
            self.load_emotion_analysis()
        elif self.model_name == "irony_analysis":
            self.load_irony_analysis()
        elif self.model_name == "sexist_analysis":
            self.load_sexist_analysis()
        elif self.model_name == "racism_analysis":
            self.load_racism_analysis()

    def get_info_about_models(self):
        info = {
            "hate_speech": {
                "function": self.load_hate_speech(),
                "types": ["robertuito", "bert"],
            },
            "incivility": {
                "function": self.load_incivility(),
                "types": ["bert"],
            },
            "toxic_speech": {
                "function": self.load_toxic_speach(),
                "types": ["political-tweets-es"],
            },
            "sentiment_analysis": {
                "function": self.load_sentiment_analysis(),
                "types": ["robertuito"],
            },
            "emotion_analysis": {
                "function": self.load_emotion_analysis(),
                "types": ["robertuito"],
            },
            "irony_analysis": {
                "function": self.load_irony_analysis(),
                "types": ["robertuito"],
            },
            "sexist_analysis": {
                "function": self.load_sexist_analysis(),
                "types": ["sexist_analysis_metwo"],
            },
            "racist_analysis": {
                "function": self.load_racism_analysis(),
                "types": ["racism_paula_lobo_et_al_average_strict"],
            },
        }
        return info

    def load_hate_speech(self, type="bert"):
        if type == "robertuito":
            self._robertuito_hate_speech_()
        elif type == "bert":
            self._bert_hate_speech_()
        else:
            raise ValueError("Unknown type of hate speech model")

    def _robertuito_hate_speech_(self):
        self.model = pipeline(
            "text-classification",
            model="pysentimiento/robertuito-hate-speech",
            truncation=True,
            max_length=128,
            device=self.device,
        )
        self.max_length = 128
        self.type_model = "hf"
        self.n_labels = 3
        self.multiclass = True
        self.labels = {
            "hateful": "hateful",
            "aggressive": "aggressive",
            "targeted": "targeted",
        }

    def _bert_hate_speech_(self):
        self.model = pipeline(
            "text-classification",
            model="jorgeortizfuentes/spanish_hate_speech",
            truncation=True,
            max_length=512,
            device=self.device,
        )
        self.max_length = 512
        self.type_model = "hf"
        self.n_labels = 2
        self.multiclass = False
        self.labels = {
            "hate": "hate",
            "no_hate": "no_hate",
        }

    def load_incivility(self, type="bert"):
        if type == "bert":
            self._bert_incivility_()
        else:
            raise ValueError("Unknown type of hate speech model")

    def _bert_incivility_(self):
        self.model = pipeline(
            "text-classification",
            model="jorgeortizfuentes/spanish_incivility",
            truncation=True,
            max_length=512,
            device=self.device,
        )
        self.max_length = 512
        self.type_model = "hf"
        self.n_labels = 2
        self.multiclass = False
        self.labels = {
            "incivility": "incivility",
            "no_incivility": "no_incivility",
        }

    def load_toxic_speach(self, type="political-tweets-es"):
        if type == "political-tweets-es":
            self._toxic_political_tweets_()
        else:
            raise ValueError("Unknown type of toxic speech model")

    def _toxic_political_tweets_(self):
        self.model = pipeline(
            "text-classification",
            model="Newtral/xlm-r-finetuned-toxic-political-tweets-es",
            truncation=True,
            max_length=512,
            device=self.device,
        )
        self.max_length = 512
        self.type_model = "hf"
        self.multiclass = False
        self.n_labels = 2
        self.labels = {
            "LABEL_0": "toxic",
            "LABEL_1": "very_toxic",
        }

    def load_sentiment_analysis(self, type="robertuito"):
        if type == "robertuito":
            self._robertuito_sentiment_analysis_()
        else:
            raise ValueError("Unknown type of sentiment analysis model")

    def _robertuito_sentiment_analysis_(self):
        self.model = pipeline(
            "text-classification",
            model="pysentimiento/robertuito-sentiment-analysis",
            truncation=True,
            max_length=128,
            device=self.device,
        )
        self.max_length = 128
        self.type_model = "hf"
        self.multiclass = False
        self.n_labels = 3
        self.labels = {"NEU": "neutral", "NEG": "negative", "POS": "positive"}

    def load_emotion_analysis(self, type="robertuito"):
        if type == "robertuito":
            self._robertuito_emotion_analysis_()
        else:
            raise ValueError("Unknown type of sentiment analysis model")

    def _robertuito_emotion_analysis_(self):
        self.model = pipeline(
            "text-classification",
            model="pysentimiento/robertuito-emotion-analysis",
            truncation=True,
            max_length=128,
            device=self.device,
        )
        self.max_length = 128
        self.type_model = "hf"
        self.multiclass = False
        self.n_labels = 7
        self.labels = {
            "others": "others",
            "joy": "joy",
            "sadness": "sadness",
            "anger": "anger",
            "surprise": "surprise",
            "disgust": "disgust",
            "fear": "fear",
        }

    def load_irony_analysis(self, type="robertuito"):
        if type == "robertuito":
            self._robertuito_irony_analysis_()
        else:
            raise ValueError("Unknown type of irony analysis model")

    def _robertuito_irony_analysis_(self):
        self.model = pipeline(
            "text-classification",
            model="pysentimiento/robertuito-irony",
            truncation=True,
            max_length=128,
            device=self.device,
        )
        self.max_length = 128
        self.type_model = "hf"
        self.multiclass = False
        self.n_labels = 2
        self.labels = {"not ironic": "not_ironic", "ironic": "ironic"}

    def load_sexist_analysis(self, type="sexist_analysis_metwo"):
        if type == "sexist_analysis_metwo":
            self._sexist_analysis_metwo_()
        else:
            raise ValueError("Unknown type of sexist analysis model")

    def _sexist_analysis_metwo_(self):
        self.model = pipeline(
            "text-classification",
            model="hackathon-pln-es/twitter_sexismo-finetuned-exist2021-metwo",
            truncation=True,
            max_length=128,
            device=self.device,
        )
        self.max_length = 128
        self.type_model = "hf"
        self.multiclass = False
        self.n_labels = 2
        self.labels = {"LABEL_0": "not_sexist", "LABEL_1": "sexist"}

    def load_racism_analysis(
        self, type="racism_paula_lobo_et_al_average_strict"
    ):
        if type == "racism_paula_lobo_et_al_average_strict":
            self._racism_paula_lobo_et_al_average_()
        else:
            raise ValueError("Unknown type of racism analysis model")

    def _racism_paula_lobo_et_al_average_(self):
        self.model = pipeline(
            "text-classification",
            model="MartinoMensio/racism-models-m-vote-strict-epoch-4",
            tokenizer="dccuchile/bert-base-spanish-wwm-uncased",
            truncation=True,
            max_length=512,
            device=self.device,
        )
        self.max_length = 512
        self.type_model = "hf"
        self.multiclass = False
        self.n_labels = 2
        self.labels = {"non-racist": "non-racist", "racist": "racist"}

    def _predict_hf_(self, text):
        prediction = self.model(
            text,
            top_k=self.n_labels,
            truncation=True,
            max_length=self.max_length,
        )
        d_prediction = {}
        for p in prediction:
            # p["label"] = self.labels[p["label"]]
            d_prediction[self.labels[p["label"]]] = p["score"]
        return d_prediction

    def predict(self, text):
        if self.type_model == "hf":
            if isinstance(text, str):
                self.last_prediction = self._predict_hf_(text)
                return self.last_prediction
            elif isinstance(text, list):
                self.last_prediction = []
                for t in text:
                    self.last_prediction.append(self._predict_hf_(t))
                return self.last_prediction



import os

DETERMINANTES = [
    "el",
    "la",
    "los",
    "las",
    "un",
    "una",
    "unos",
    "unas",
    "esta",
    "estas",
    "este",
    "estos",
    "esa",
    "esas",
    "ese",
    "esos",
    "aquel",
    "aquella",
    "aquellas",
    "aquellos",
    "mi",
    "tu",
    "su",
    "mis",
    "tus",
    "sus",
    "nuestro",
    "nuestra",
    "nuestros",
    "nuestras",
    "vuestro",
    "vuestra",
    "vuestros",
    "vuestras",
    "su",
    "sus",
    "cuyo",
    "cuya",
    "cuyos",
    "cuyas",
    "cuánto",
    "cuántos",
    "cuánta",
    "cuántas",
    "qué",
    "tan",
    "menos",
    "más",
    "algún",
    "alguna",
    "algunos",
    "algunas",
    "ningún",
    "ninguna",
    "ningunos",
    "ningunas",
]

PREPOSICIONES = [
    "a",
    "ante",
    "bajo",
    "cabe",
    "con",
    "contra",
    "de",
    "desde",
    "durante",
    "en",
    "entre",
    "hacia",
    "hasta",
    "mediante",
    "para",
    "por",
    "según",
    "sin",
    "so",
    "sobre",
    "tras",
    "versus",
    "vía",
]

PRONOMBRES = [
    "yo",
    "tú",
    "vos",
    "vo",
    "él",
    "ella",
    "ello",
    "elle",
    "usted",
    "nosotros",
    "nosotras",
    "nosotres",
    "vosotros",
    "vosotras",
    "vosotres",
    "ellos",
    "ellas",
    "ustedes",
    "elles",
    "mí",
    "ti",
    "sí",
    "consigo",
]

CONJUNCIONES = [
    "y",
    "ni",
    "sino",
    "tanto",
    "como",
    "que",
    "pero",
    "mas",
    "empero",
    "mientras",
    "o",
    "bien",
    "ya",
]

default_stopwords = list(set(DETERMINANTES + PREPOSICIONES + PRONOMBRES + CONJUNCIONES))

extended_stopwords = [
    "cómo",
    "propia",
    "por qué",
    "será",
    "uso",
    "cinco",
    "hubiera",
    "usa",
    "emplean",
    "todavía",
    "mal",
    "dan",
    "quiénes",
    "sólo",
    "tú",
    "para",
    "hicieron",
    "podriais",
    "apenas",
    "aquello",
    "manera",
    "trabaja",
    "arriba",
    "fueran",
    "tus",
    "unos",
    "está",
    "haces",
    "muchas",
    "proximo",
    "ellas",
    "aún",
    "aquella",
    "no",
    "has",
    "hoy",
    "delante",
    "hubiese",
    "encuentra",
    "mediante",
    "seamos",
    "suyos",
    "tendríais",
    "través",
    "mucho",
    "mas",
    "toda",
    "ciertos",
    "habrías",
    "nada",
    "bien",
    "ello",
    "son",
    "su",
    "varios",
    "éramos",
    "detrás",
    "tendríamos",
    "eran",
    "nadie",
    "poco",
    "aun",
    "conseguimos",
    "pais",
    "mi",
    "estuviste",
    "intentar",
    "como",
    "decir",
    "ellos",
    "estuvo",
    "claro",
    "tambien",
    "porque",
    "hacer",
    "fuera",
    "última",
    "ningunos",
    "seré",
    "primera",
    "tuyo",
    "tendrán",
    "éstos",
    "estaréis",
    "aqui",
    "han",
    "tendrías",
    "cuando",
    "vuestras",
    "fueseis",
    "tenga",
    "una",
    "sabemos",
    "hago",
    "otro",
    "largo",
    "aquellas",
    "mis",
    "deprisa",
    "míos",
    "empleas",
    "hecho",
    "mayor",
    "allí",
    "estuvisteis",
    "estábamos",
    "ocho",
    "habidas",
    "hacemos",
    "queremos",
    "propio",
    "mientras",
    "estuvieseis",
    "u",
    "estaba",
    "quizá",
    "grandes",
    "qué",
    "usamos",
    "breve",
    "siempre",
    "muy",
    "ojalá",
    "nuestras",
    "actualmente",
    "hacen",
    "aquéllas",
    "días",
    "hubierais",
    "últimas",
    "estaría",
    "seréis",
    "dicho",
    "habíamos",
    "gueno",
    "momento",
    "es",
    "bajo",
    "tendría",
    "estaremos",
    "lejos",
    "estemos",
    "ojala",
    "nuevas",
    "lugar",
    "todavia",
    "nunca",
    "estaré",
    "estabas",
    "ése",
    "tuya",
    "he",
    "mios",
    "indicó",
    "tengo",
    "pues",
    "hace",
    "podrá",
    "me",
    "nuestra",
    "hacerlo",
    "despues",
    "quizás",
    "estadas",
    "ésta",
    "estuviesen",
    "intentais",
    "los",
    "tendréis",
    "tu",
    "voy",
    "además",
    "sola",
    "tengas",
    "tienes",
    "estáis",
    "general",
    "último",
    "estan",
    "soyos",
    "habréis",
    "enfrente",
    "empleo",
    "si",
    "ésos",
    "aquélla",
    "aseguró",
    "buena",
    "ex",
    "tenéis",
    "mencionó",
    "sus",
    "ir",
    "señaló",
    "arribaabajo",
    "cuales",
    "da",
    "ver",
    "debajo",
    "podeis",
    "eras",
    "ante",
    "éstas",
    "casi",
    "contra",
    "existe",
    "parece",
    "tenemos",
    "fuiste",
    "hube",
    "existen",
    "habido",
    "tuvieses",
    "quedó",
    "todo",
    "ninguno",
    "hay",
    "sea",
    "estais",
    "eso",
    "ahí",
    "pocas",
    "hemos",
    "temprano",
    "expresó",
    "muchos",
    "hubimos",
    "pronto",
    "hubo",
    "solamente",
    "ha",
    "teneis",
    "cuenta",
    "estuvimos",
    "estando",
    "h",
    "segunda",
    "suyas",
    "fue",
    "dice",
    "tendrás",
    "valor",
    "hubieseis",
    "seis",
    "incluso",
    "q",
    "serás",
    "atras",
    "sera",
    "habían",
    "suyo",
    "habré",
    "o",
    "estado",
    "cuánto",
    "sois",
    "fuese",
    "tenida",
    "diferente",
    "podria",
    "hayamos",
    "aunque",
    "sabeis",
    "nuestro",
    "quizas",
    "tener",
    "estas",
    "consiguen",
    "tuve",
    "trabajamos",
    "ambos",
    "les",
    "solas",
    "g",
    "ampleamos",
    "antaño",
    "anterior",
    "propios",
    "consideró",
    "despacio",
    "tuvieseis",
    "poner",
    "posible",
    "medio",
    "le",
    "tampoco",
    "junto",
    "asi",
    "vaya",
    "primer",
    "cuántas",
    "ayer",
    "fui",
    "cuanta",
    "cuántos",
    "tenía",
    "cuantas",
    "hasta",
    "respecto",
    "demasiado",
    "la",
    "varias",
    "luego",
    "debe",
    "tenidas",
    "también",
    "dias",
    "otros",
    "modo",
    "primeros",
    "d",
    "mio",
    "vez",
    "segundo",
    "estuve",
    "dos",
    "tuvieron",
    "ahora",
    "saber",
    "dónde",
    "tened",
    "tres",
    "mías",
    "al",
    "alrededor",
    "todos",
    "sin",
    "ni",
    "dieron",
    "llevar",
    "seríais",
    "sé",
    "ti",
    "trabajas",
    "todas",
    "antano",
    "habrían",
    "hablan",
    "hubieras",
    "hacia",
    "emplear",
    "habremos",
    "hubiéramos",
    "dijo",
    "excepto",
    "mío",
    "sido",
    "cerca",
    "cierto",
    "quién",
    "hubieran",
    "intento",
    "tenidos",
    "estéis",
    "ese",
    "n",
    "lo",
    "tuviera",
    "bueno",
    "intentas",
    "habia",
    "comentó",
    "manifestó",
    "por",
    "quien",
    "algo",
    "estarán",
    "parte",
    "sería",
    "haciendo",
    "habríamos",
    "estarás",
    "cuáles",
    "habrán",
    "realizar",
    "tenido",
    "intentan",
    "sigue",
    "estaríamos",
    "fuésemos",
    "ninguna",
    "añadió",
    "menudo",
    "tuvisteis",
    "sino",
    "mia",
    "fuisteis",
    "estados",
    "tercera",
    "eres",
    "estás",
    "tarde",
    "raras",
    "estuvieras",
    "uno",
    "dejó",
    "tengan",
    "habías",
    "van",
    "vosotras",
    "estabais",
    "ahi",
    "nuevos",
    "sabe",
    "seáis",
    "dar",
    "nosotras",
    "estamos",
    "trabajo",
    "estar",
    "tuviste",
    "tenías",
    "alguna",
    "considera",
    "yo",
    "consigues",
    "tuvimos",
    "están",
    "mucha",
    "eramos",
    "pueda",
    "fuéramos",
    "nosotros",
    "creo",
    "esas",
    "tendré",
    "nuestros",
    "este",
    "principalmente",
    "usar",
    "ya",
    "aproximadamente",
    "poca",
    "dentro",
    "tendremos",
    "dijeron",
    "habiendo",
    "aquel",
    "día",
    "entonces",
    "entre",
    "f",
    "fin",
    "mí",
    "trabajais",
    "salvo",
    "tuvierais",
    "solo",
    "trata",
    "intentamos",
    "buen",
    "buenas",
    "usais",
    "últimos",
    "después",
    "primero",
    "tuviésemos",
    "verdadero",
    "conmigo",
    "otra",
    "trabajar",
    "serían",
    "estará",
    "ella",
    "estuvieses",
    "según",
    "vais",
    "aquellos",
    "tuviese",
    "próximo",
    "esos",
    "peor",
    "diferentes",
    "de",
    "menos",
    "buenos",
    "del",
    "suya",
    "aquél",
    "algún",
    "teniendo",
    "podriamos",
    "veces",
    "vosotros",
    "éste",
    "antes",
    "estuviésemos",
    "horas",
    "dicen",
    "igual",
    "erais",
    "realizó",
    "habéis",
    "hubiésemos",
    "mias",
    "más",
    "segun",
    "fueras",
    "b",
    "cierta",
    "vuestra",
    "con",
    "quiza",
    "solos",
    "teníais",
    "habida",
    "enseguida",
    "tiene",
    "fuesen",
    "sal",
    "tuvieran",
    "tendrá",
    "esta",
    "fueron",
    "estad",
    "total",
    "misma",
    "donde",
    "esa",
    "aquí",
    "pocos",
    "estos",
    "a",
    "tuviesen",
    "vuestro",
    "dio",
    "paìs",
    "soy",
    "haber",
    "seremos",
    "va",
    "somos",
    "dia",
    "intenta",
    "conocer",
    "lleva",
    "hayas",
    "os",
    "podrían",
    "habrás",
    "él",
    "tuvo",
    "estoy",
    "cosas",
    "repente",
    "mía",
    "conseguir",
    "usas",
    "hayáis",
    "tanto",
    "tienen",
    "pasada",
    "aquéllos",
    "estarías",
    "mejor",
    "desde",
    "seríamos",
    "tuvieras",
    "cuándo",
    "informo",
    "próximos",
    "era",
    "estuvierais",
    "estés",
    "seas",
    "puede",
    "empleais",
    "hubieses",
    "hubiesen",
    "podemos",
    "serías",
    "detras",
    "durante",
    "podrán",
    "cuantos",
    "mismos",
    "ésas",
    "pasado",
    "sean",
    "vamos",
    "cada",
    "habrá",
    "las",
    "haya",
    "estuvieran",
    "hubieron",
    "sabes",
    "tengáis",
    "pueden",
    "consigo",
    "podrias",
    "tenían",
    "cualquier",
    "hubiste",
    "vuestros",
    "ciertas",
    "pesar",
    "xq",
    "había",
    "tiempo",
    "ningún",
    "serán",
    "adelante",
    "gran",
    "esto",
    "ademas",
    "verdad",
    "verdadera",
    "cuál",
    "tengamos",
    "c",
    "pudo",
    "esté",
    "cual",
    "estada",
    "mismo",
    "ningunas",
    "sí",
    "debido",
    "estuvieron",
    "podrian",
    "estén",
    "partir",
    "poder",
    "estaban",
    "tan",
    "dado",
    "cuanto",
    "se",
    "ustedes",
    "nueva",
    "el",
    "otras",
    "ejemplo",
    "llegó",
    "usted",
    "tuyos",
    "lado",
    "fueses",
    "podría",
    "alli",
    "sobre",
    "mismas",
    "nos",
    "embargo",
    "ultimo",
    "teníamos",
    "tras",
    "que",
    "hubisteis",
    "habría",
    "supuesto",
    "cuatro",
    "habla",
    "encima",
    "estaríais",
    "contigo",
    "estuviera",
    "tendrían",
    "habíais",
    "tuviéramos",
    "usan",
    "adrede",
    "estarían",
    "quiere",
    "un",
    "informó",
    "demás",
    "habríais",
    "algunas",
    "hizo",
    "pero",
    "tuyas",
    "v",
    "deben",
    "así",
    "saben",
    "ser",
    "te",
    "nuevo",
    "consigue",
    "hayan",
    "bastante",
    "siete",
    "fuerais",
    "estuviéramos",
    "tal",
    "qeu",
    "siguiente",
    "propias",
    "trabajan",
    "cuánta",
    "algunos",
    "explicó",
    "haceis",
    "quienes",
    "realizado",
    "habidos",
    "en",
    "puedo",
    "siendo",
    "estuviese",
    "final",
    "alguno",
    "unas",
    "fuimos",
    "ésa",
    "ojalá",
]

extended_stopwords = list(set(default_stopwords + extended_stopwords))

punct = [
    ".",
    ",",
    ":",
    ";",
    "!",
    "?",
    "¿",
    "¡",
    "(",
    ")",
    "[",
    "]",
    "{",
    "}",
    "/",
    "\\",
    "|",
    "-",
    "_",
    "–",
    "—",
    "…",
    "·",
    "•",
    "°",
    "´",
    "`",
    "'",
    '"',
    "“",
    "”",
    "‘",
    "’",
    "«",
    "»",
    "‹",
    "›",
]


import numpy as np
import math
import torch
from transformers import pipeline, AutoTokenizer
import warnings
from spanish_nlp.utils.stopwords import punct
from .abstract import DataAugmentationAbstract
import re

from transformers.utils import logging
logging.set_verbosity(40)


class Masked(DataAugmentationAbstract):
    """
    Class for data augmentation with spelling. It uses the Huggingface pipeline for masked language modeling.

    """

    def __init__(
        self,
        method="sustitute",
        model="dccuchile/bert-base-spanish-wwm-cased",
        tokenizer="default",
        stopwords="default",
        aug_percent=0.4,
        device=None,
        top_k=10,
    ):
        """Init class.

        Args:
            model (str, optional): huggingface model. Defaults to "dccuchile/bert-base-spanish-wwm-cased".
            tokenizer (str, optional): huggingface tokenizer. Defaults to "default". If "default" is used, the model is used as tokenizer.
            stopwords (list or str, optional): list of stopwords. Defaults to "default". If "default" is used, the default stopwords are used.
            aug_percent (float, optional): percentage of words to be augmented. Defaults to 0.1.
            device (str or torch.device, optional): device to use. Defaults to None. If None is used, the device is automatically selected.
            top_k (int, optional): number of top k words to be used for augmentation. Defaults to 10.
        """
        super().__init__(method)

        if self.method not in ["sustitute", "insert"]:
            raise ValueError("The method must be 'sustitute' or 'insert'")

        # Set the parameters
        self.aug_percent = aug_percent
        self.tokenizer = tokenizer
        self.stopwords = stopwords
        self.top_k = top_k

        if self.stopwords == "default":
            self._load_default_stopwords_()

        # Download Huggingface model, tokenizer, device and mask token
        if device is None:
            # All available gpu devices if any else cpu
            if torch.cuda.is_available():
                self.device = torch.device("cuda")
            else:
                self.device = torch.device("cpu")
        else:
            self.device = device
        if self.tokenizer == "default":
            self.tokenizer = model
        self.__load__tokenizer__()
        self.fillmask = pipeline(
            "fill-mask", model=model, tokenizer=self.tokenizer, device=self.device
        )
        self.mask_token = self.fillmask.tokenizer.mask_token

    def __load__tokenizer__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(self.tokenizer,
                                                       truncation=True,
                                                       use_fast=True)

    def set_aug_percent(self, aug_percent):
        """
        Set the augmentation percentage.
        """
        if not isinstance(aug_percent, float):
            raise ValueError("The aug_percent must be a float.")
        self.aug_percent = aug_percent

    def _text_augment_(self, text, num_samples=1):
        """
        Augment a single text.
        """
        if self.method == "sustitute":
            return self._sustitute_augment_(text, num_samples)
        elif self.method == "insert":
            return self._insert_augment_(text, num_samples)
        else:
            raise ValueError("The method must be 'sustitute' or 'insert'")

    def _sustitute_augment_(self, text, num_samples=1):
        """
        Augment a text with the sustitute method. This method replaces a percentage of the words with other words.

        Args:
            text (str): text to be augmented.
            num_samples (int, optional): number of samples to be generated. Defaults to 1.

        Returns:
            list: list of augmented texts.
        """

        output_texts = []
        for _ in range(num_samples):
            # Tokenize text, count the tokens and if the tokens > max_length, return the original sentence
            n_tokens = len(self.tokenizer.tokenize(text))
            n_tokens_split = int(
                self.tokenizer.model_max_length / (self.aug_percent+1))+1
            n_splits = math.ceil((n_tokens / n_tokens_split))

            if n_splits > 1:
                new_text = self._large_sustitute_(text)
            else:
                new_text = self._normal_sustitute_(text)

            if new_text not in output_texts:
                output_texts.append(new_text)
            else:
                num_samples += 1

        return output_texts

    def _normal_sustitute_(self, sentence, num_samples=1):
        """
        Augments the input sentence by replacing a percentage of its non-stopword words with predicted words obtained by
        masking them using a pre-trained language model.

        Args:
            sentence (str): Input sentence to be augmented.
        Returns:
            str: Augmented sentence.
        """
        words = sentence.split(" ")
        not_allowed = punct + self.stopwords

        n_stopwords = sum([1 for word in words if word in not_allowed])
        n_total_words = len(words) - n_stopwords

        num_words = int(n_total_words * self.aug_percent)
        K_list = []

        # Select the words to be replaced
        while len(K_list) < num_words:
            K = np.random.randint(0, len(words))
            if words[K] not in punct:
                K_list.append(K)

        # Iterate over the words to be replaced
        for K in K_list:
            words = sentence.split(" ")
            masked_sentence = " ".join(
                words[:K] + [self.mask_token] + words[K + 1:])
            predictions = self.fillmask(masked_sentence, top_k=self.top_k)
            random_number = np.random.randint(0, self.top_k)
            new_word = predictions[random_number]["token_str"]

            # Verify that the new word is not a punctuation or a stopword
            count = 0
            while True:
                random_number = np.random.randint(0, self.top_k)
                # Check if there is a punctuation in new_word
                pattern = "|".join(re.escape(p) for p in punct)
                if re.search(pattern, new_word) == None and new_word not in not_allowed:
                    break
                elif count > self.top_k:
                    break
                else:
                    count += 1
            # Save the new sentence
            sentence = predictions[random_number]["sequence"]
        return sentence

    def _large_sustitute_(self, sentence, num_samples=1):
        """
        Augments the input sentence by replacing a percentage of its non-stopword words with predicted words obtained by
        masking them using a pre-trained language model. This method is used for long texts that exceed the maximum token
        length allowed by the language model.

        Args:
            sentence (str): Input sentence to be augmented.
        Returns:
            str: Augmented sentence.
        """
        # Split the text into chunks with the desired length
        n_tokens = len(self.tokenizer.tokenize(sentence))
        n_tokens_split = int(
            self.tokenizer.model_max_length / (self.aug_percent+1)) + 1
        n_splits = math.ceil((n_tokens / n_tokens_split))
        sentence_splits = self._split_text_into_chunks_(
            sentence, n_splits=n_splits)

        # Augment each chunk of the text and concatenate the results
        augmented_chunks = []
        for chunk in sentence_splits:
            augmented_chunk = self._normal_sustitute_(chunk, num_samples=1)
            augmented_chunks.append(augmented_chunk)

        augmented_sentence = " ".join(augmented_chunks)

        return augmented_sentence

    def _split_text_into_chunks_(self, text, n_splits):
        """
        Splits the input text into n_splits chunks with approximately the same length.

        Args:
            text (str): Input text to be split.
            n_splits (int): Number of chunks to split the text into.

        Returns:
            list of str: List of chunks.
        """
        tokens = self.tokenizer.tokenize(text)
        n_tokens = len(tokens)
        chunk_size = int(n_tokens / n_splits)
        chunks = [tokens[i:i+chunk_size]
                  for i in range(0, n_tokens, chunk_size)]
        return [" ".join(chunk) for chunk in chunks]

    def _insert_augment_(self, text, num_samples=1, max_words=450):
        """
        Augment a text with the insert method. This method inserts new words in the text.

        Args:
            text (str): text to be augmented.
            num_samples (int, optional): number of samples to be generated. Defaults to 1.

        Returns:
            list: list of augmented texts.
        """
        output_texts = []
        for _ in range(num_samples):
            # Tokenize text, count the tokens and if the tokens > max_length, return the original sentence
            n_tokens = len(self.tokenizer.tokenize(text))
            n_tokens_split = int(
                self.tokenizer.model_max_length / (self.aug_percent+1))+1
            n_splits = math.ceil((n_tokens / n_tokens_split))

            if n_splits > 1:
                new_text = self._large_insert_(text)
            else:
                new_text = self._normal_insert_(text)

            if new_text not in output_texts:
                output_texts.append(new_text)
            else:
                num_samples += 1
        return output_texts

    def _normal_insert_(self, sentence, num_samples=1):
        """Insert a self.aug_percent% of words within the sentence (not after punctuation marks)

        Args:
            sentence (str): Input sentence to be augmented.
        Returns:
            str: Augmented sentence.
        """

        # Tokens that are not allowed to be replaced
        words = sentence.split(" ")
        not_allowed = punct + self.stopwords
        # Tokenize text, count the tokens and if the tokens > max_length, return the original sentence
        tokens = len(self.tokenizer.tokenize(sentence))
        num_words = int(len(words) * self.aug_percent)
        K_list = []

        # Select the words to be replaced
        while len(K_list) < num_words:
            K = np.random.randint(0, len(words))
            if words[K] not in not_allowed:
                K_list.append(K)

        # Iterate over the words to be replaced
        for K in K_list:
            words = sentence.split(" ")
            masked_sentence = " ".join(
                words[:K] + [self.mask_token] + words[K:])
            predictions = self.fillmask(masked_sentence,
                                        top_k=self.top_k)
            random_number = np.random.randint(0, self.top_k)
            new_word = predictions[random_number]["token_str"]

            # Verify that the new word is not a punctuation or a stopword
            count = 0
            while True:
                random_number = np.random.randint(0, self.top_k)
                pattern = "|".join(re.escape(p) for p in punct)
                if re.search(pattern, new_word) == None and new_word not in not_allowed:
                    break
                elif count > self.top_k:
                    break
                else:
                    count += 1
            # Save the new sentence
            sentence = predictions[random_number]["sequence"]
        return sentence

    def _large_insert_(self, sentence, num_samples=1):
        """
        Augment a text with the insert method. This method inserts new words to the text.

        Args:
            sentence (str): Input sentence to be augmented.
        Returns:
            str: Augmented sentence.
        """
        # Split the text into chunks with the desired length
        n_tokens = len(self.tokenizer.tokenize(sentence))
        n_tokens_split = int(
            self.tokenizer.model_max_length / (self.aug_percent+1)) + 1
        n_splits = math.ceil((n_tokens / n_tokens_split))
        sentence_splits = self._split_text_into_chunks_(
            sentence, n_splits=n_splits)

        # Augment each chunk of the text and concatenate the results
        augmented_chunks = []
        for chunk in sentence_splits:
            augmented_chunk = self._normal_insert_(chunk, num_samples=1)
            augmented_chunks.append(augmented_chunk)

        augmented_sentence = " ".join(augmented_chunks)
        return augmented_sentence